//----------------------------------------------------------------------------------
// dݓǂݍ݃NX
// Copyright (c) 2008,2009 N099
//----------------------------------------------------------------------------------
#include <fstream>
#include <new>
#include <cmath>

#include <string>
#include <vector>
#include <map>
#include <algorithm>

#include "weight_size.h"
#include "weight_reader.h"

using namespace std;

///WeightReader
WeightReader::WeightReader() : BaseWeightReader("?")
{
}

WeightReader::~WeightReader()
{
}

bool WeightReader::LoadFile(const char *name)
{
	ifstream ifs(name);

	if(!ifs){
		return false;
	}

	char buf[1024];
	while(!ifs.eof()){
		ifs.getline(buf, sizeof(buf));
		if(buf[0] == '#'){ //Rg
			continue;
		}else if(buf[0] == '$'){ //O
			_name = &buf[1];
		}else{
			buf[sizeof(buf) - 1] = '\0';
			AppendWeight(buf);
		}
	}

	ifs.close();
	if(_radius.empty()){
		return false;
	}else{
		_file_name = name;
		return true;
	}
}

int WeightReader::GetWeight(int r, vector<float> &g) const
{
	if(r < 1 || r > static_cast<int>(_radius.size())){
		return -1;
	}
	--r; //r"+1"ItZbgĂ

	const int ww = WIDTH_WEIGHT - 1;
	int rd = _radius[r] % ww;
	if(rd){
		rd = (_radius[r] / ww) + 1;
	}else{
		rd = _radius[r] / ww;
		rd = (rd > 0) ? rd : 1;
	}
	g.resize(rd * WIDTH_WEIGHT);

	int pos = 0;
	const map<int, float> &wt = _weight[r];
	vector<float>::iterator itr = g.begin();

	for(int i=0; i<rd; ++i){
		map<int, float>::const_iterator ct = wt.end();
		for(int j=0; j<ww; ++j){
			ct = wt.find(pos);
			if(ct != wt.end()){ //found
				*itr++ = ct->second;
			}else{ //not set
				*itr++ = (pos > 0) ? 0.0f : 1.0f;
			}
			pos++;
		}
		//n*(WIDTH_WEIGHT-1)  n*(WIDTH_WEIGHT) ɂ͓l
		*itr++ = (ct != wt.end()) ? ct->second : 0.0f;
	}

	return rd;
}


void WeightReader::AppendWeight(char *buf)
{
	typedef pair<int, float> Weight_Pair;

	char *head = buf;
	while(*buf != '\0'){
		if(*buf == '"' || *buf == ','){
			*buf = ' ';
		}
		++buf;
	}

	int r = -1, i = -1;
	float f = 0.0f;
	int ret = sscanf_s(head, "%d %d %f", &r, &i, &f);
	if(ret != 3 || r < 0 || i < 0 || f == 0.0f){
		return;
	}else if(r >= _set_max || i >= (HEIGHT_WEIGHT-1) * (WIDTH_WEIGHT-1)){
		//d݃eNX`TCYɂ鐧
		return;
	}

	if(static_cast<int>(_radius.size()) < (r + 1)){
		_radius.resize(r + 1);
		_weight.resize(r + 1);
	}

	_radius[r] = max(_radius[r], i);
	_weight[r].insert(Weight_Pair(i, f));
}

//for Lanczos windowed sinc
class WeightLanczos : public BaseWeightReader {
public :
	WeightLanczos() : BaseWeightReader("Lanczos") {}
	int GetWeight(int r, vector<float> &g) const;
	int GetRadiusNum() const { return HEIGHT_WEIGHT - 1; }
};

int WeightLanczos::GetWeight(int r, vector<float> &g) const
{
	if(r <= 0 || r >= HEIGHT_WEIGHT){
		return -1;
	}
	g.resize(r * WIDTH_WEIGHT);

	const int ww = WIDTH_WEIGHT - 1;
	vector<float>::iterator itr = g.begin();

	for(int j=0; j<r; ++j){
		*itr++ = (j == 0) ? 1.0f : 0.0f;
		for(int i=1; i<WIDTH_WEIGHT; i++){
			double d = (M_PI * (i + j * ww))/ww;
			double dr2 = r/(d * d);
			d = dr2 * sin(d) * sin(d/r);
			*itr++ = static_cast<float>(d);
		}
	}

	return r;
}

//for Rectangle window
class WeightRect : public BaseWeightReader {
public :
	WeightRect() : BaseWeightReader("Point") {}
	int GetWeight(int r, vector<float> &g) const;
	int GetRadiusNum() const { return 1; }
};

int WeightRect::GetWeight(int, vector<float> &g) const
{
	g.resize(WIDTH_WEIGHT);

	int i = 0;
	for( ; i<WIDTH_WEIGHT/2; ++i) g[i] = 1.0f;
	if((WIDTH_WEIGHT-1)%2 == 0) g[i++] = 0.5f; //0.5Tv_
	for( ; i<WIDTH_WEIGHT; ++i) g[i] = 0.0f;

	return 1;
}

//for Tent window
class WeightTent : public BaseWeightReader {
public :
	WeightTent() : BaseWeightReader("Linear") {}
	int GetWeight(int r, vector<float> &g) const;
	int GetRadiusNum() const { return 1; }
};

int WeightTent::GetWeight(int, vector<float> &g) const
{
	g.resize(WIDTH_WEIGHT);

	const float r = 1.0f / (WIDTH_WEIGHT - 1);
	for(int i=0; i<WIDTH_WEIGHT; ++i){
		g[i] = 1.0f - i * r;
	}

	return 1;
}

//RectTent ̒
class WeightRT : public BaseWeightReader {
public :
	WeightRT() : BaseWeightReader("Point/Linear") {}
	int GetWeight(int r, vector<float> &g) const;
	int GetRadiusNum() const { return 17; }
};

int WeightRT::GetWeight(int r, vector<float> &g) const
{
	r--; //r = [0,16]
	g.resize(WIDTH_WEIGHT);

	if(r == 0){ //Rect
		int i = 0;
		for( ; i<WIDTH_WEIGHT/2; ++i) g[i] = 1.0f;
		if((WIDTH_WEIGHT-1)%2 == 0) g[i++] = 0.5f;
		for( ; i<WIDTH_WEIGHT; ++i) g[i] = 0.0f;
	}else if(r == GetRadiusNum() - 1){ //Tent
		float d = 1.0f / (WIDTH_WEIGHT - 1);
		for(int i=0; i<WIDTH_WEIGHT; ++i){
			g[i] = 1.0f - i * d;
		}
	}else{
		r *= WIDTH_WEIGHT/(2 * (GetRadiusNum() - 1));
		r += WIDTH_WEIGHT/2;
		float d = 1.0f / (WIDTH_WEIGHT - 1);
		float a = d * r;
		float b = 1.0f / (1.0f - 2.0f * a);
		for(int i=0; i<WIDTH_WEIGHT; ++i){
			g[i] = min(1.0f, max(0.0f, b * (d * i - a)));
		}
	}

	return 1;
}

// Mitchell-Netravali filter
//  ref.
//   Mitchell, D. P. and Netravali, A. N. 1988.
//   Reconstruction filters in computer-graphics.
//   SIGGRAPH Comput. Graph. 22, 4 (Aug. 1988), 221-228.
//a
class WeightPWCubic : public BaseWeightReader {
public :
	WeightPWCubic() : BaseWeightReader("Cubic") {}
	int GetWeight(int r, vector<float> &g) const;
	int GetRadiusNum() const { return 13; }
};

int WeightPWCubic::GetWeight(int r, vector<float> &g) const
{
	g.resize(WIDTH_WEIGHT*2);

	float x3,x2,x;
	float b = (r - 1) / (float)(GetRadiusNum() - 1); //r = [1,13]
	float c = 0.5f * (1.0f - b);
	const float d = 1.0f / (WIDTH_WEIGHT - 1);

	// 0 < |x| < 1
	for(int i=0; i<WIDTH_WEIGHT; ++i){
		x = d * i;
		x2 = x * x;
		x3 = x2 * x;
		g[i] = ((12.0f - 9.0f * b - 6.0f * c) * x3
			+ (-18.0f + 12.0f * b + 6.0f * c) * x2 + (6.0f - 2.0f * b)) / 6.0f;
	}
	// 1 < |x| < 2
	for(int i=0; i<WIDTH_WEIGHT; ++i){
		x = d * (i + WIDTH_WEIGHT - 1);
		x2 = x * x;
		x3 = x2 * x;
		g[i + WIDTH_WEIGHT] =
			((-b - 6.0f * c) * x3 + (6.0f * b + 30.0f * c) * x2
			+ (-12.0f * b - 48.0f * c) * x + (8.0f * b + 24.0f * c)) / 6.0f;
	}

	return 2;
}

//BaseWeightReader
BaseWeightReader::BaseWeightReader(const char *name) : _name(name)
{
}

BaseWeightReader::~BaseWeightReader()
{
}

bool BaseWeightReader::LoadWeight(const char *name, BaseWeightReader* &wr)
{
	if(strcmp(name, "<Lanczos>") == 0){
		wr = new(nothrow) WeightLanczos;
	}else if(strcmp(name, "<Rect>") == 0){
		wr = new(nothrow) WeightRect;
	}else if(strcmp(name, "<Tent>") == 0){
		wr = new(nothrow) WeightTent;
	}else if(strcmp(name, "<Rect&Tent>") == 0){
		wr = new(nothrow) WeightRT;
	}else if(strcmp(name, "<PWCubic>") == 0){
		wr = new(nothrow) WeightPWCubic;
	}else{
		wr = NULL;
		WeightReader *tmp = new(nothrow) WeightReader;
		if(tmp != NULL){
			if(tmp->LoadFile(name)){ //read weight file
				wr = tmp;
			}else{
				delete tmp; //failed
			}
		}
	}

	return (wr == NULL) ? false : true;
}
